// ata.c:  ATA simulator for vblade
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include "dat.h"
#include "fns.h"

enum {
	// err bits
	UNC =	1<<6,
	MC =	1<<5,
	IDNF =	1<<4,
	MCR =	1<<3,
	ABRT = 	1<<2,
	NM =	1<<1,

	// status bits
	BSY =	1<<7,
	DRDY =	1<<6,
	DF =		1<<5,
	DRQ =	1<<3,
	ERR =	1<<0,
};

static ushort ident[256] = {
	[47] 0x8000,
	[49] 0x0200,
	[50] 0x4000,
	[83] 0x5400,
	[84] 0x4000,
	[86] 0x1400,
	[87] 0x4000,
	[93] 0x400b,
};

static void
setfld(ushort *a, int idx, int len, char *str)	// set field in ident
{
	uchar *p;

	p = (uchar *)(a+idx);
	while (len > 0) {
		if (*str == 0)
			p[1] = ' ';
		else
			p[1] = *str++;
		if (*str == 0)
			p[0] = ' ';
		else
			p[0] = *str++;
		p += 2;
		len -= 2;
	}
}

static void
setlba28(ushort *p, vlong lba)
{
	p += 60;
	*p++ = lba & 0xffff;
	*p = lba >> 16 & 0x0fffffff;
}

static void
setlba48(ushort *p, vlong lba)
{
	p += 100;
	*p++ = lba;
	lba >>= 16;
	*p++ = lba;
	lba >>= 16;
	*p++ = lba;
	lba >>= 16;
	*p = lba;
}
		
void
atainit(void)
{
	char buf[64];

	setfld(ident, 27, 40, "Coraid EtherDrive vblade");
	sprintf(buf, "V%d.%d\n", 4, 0);
	setfld(ident, 23, 8, buf);
	setfld(ident, 10, 20, "SSN HERE");
}


// real programmer know how to use a goto now and then

void
atacmd(Ataregs *p, uchar *dp)		// do the ata cmd
{
	vlong lba;
	ushort *ip;
	int n;

	switch (p->cmd) {
	case 0x20:		// read sectors
		lba = p->lba & 0x00FFFFFF;	// 24 bit 
		goto rdwr;
	case 0x24:		// read sectors ext
		lba = p->lba & 0x0000ffffffffffffLL;	// full 48
		goto rdwr;
	case 0x30:		// write sectors
		lba = p->lba & 0x00FFFFFF;	// 24 bit 
		goto rdwr;
	case 0x34:		// write sectors ext
		lba = p->lba & 0x0000ffffffffffffLL;	// full 48
		goto rdwr;
	case 0xe7:		// flush cache
		return;
	case 0xec:		// identify device
		memmove(dp, ident, 512);
		ip = (ushort *)dp;
		if ((size-1) & 0xf00000)
			setlba28(ip, 0x0fffffff);
		else
			setlba28(ip, size - 1);
		setlba48(ip, size -1);
		p->err = 0;
		p->status = DRDY;
		return;
	default:
		p->status = DRDY | ERR;
		p->err = ABRT;
	}
rdwr:
	if (lba + p->sectors >= size) {
		p->err = IDNF;
		p->status = DRDY | ERR;
		p->lba = lba;
		return;
	}
	if (p->cmd == 0x20 || p->cmd == 0x24) {
		n = getsec(bfd, dp, lba, p->sectors);
		if (n == -1) {
			perror("read block device");
			p->status |= ERR;
		}
	} else {
		n = putsec(bfd, dp, lba, p->sectors);
		if (n == -1) {
			perror("write block device");
			p->status |= ERR;
		}
	}
	p->lba += p->sectors;
	p->sectors = 0;
	p->status = DRDY;
	p->err = 0;
}
